/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "gnut_lib.h"
#include "Gnut_List.h"

Gnut_List * gnut_list_unlink(Gnut_List *list, void *data, Gnut_List **unlinked)
{
  Gnut_List *tmp,*prev;
  
  g_debug(5,"list=%p data=%p\n",list,data);
  prev=NULL;
  tmp=list;
  while (tmp) {
    if (tmp->data == data) {
      if (prev) prev->next=tmp->next;
      if (list==tmp) list=list->next;
      
      *unlinked=tmp;
      break;
    }
    prev=tmp;
    tmp=tmp->next;
  }
  
  g_debug(5,"new list=%p\n",list);
  return list;
}
  

// Gnut_List * gnut_list_remove(Gnut_List *list, void *data)
//
// removes the first occurence of data in list it does NOT delete
// this memory.
// returns the pointer to the new head of the list, does not indicate
// whether the data was actually found
//
Gnut_List * gnut_list_remove(Gnut_List *list, void *data)
{
  Gnut_List *tmp,*prev;
  int flag=0;
  
  g_debug(5,"list=%p data=%p\n",list,data);
  prev=NULL;
  tmp=list;
  while (tmp) {
//    g_debug(5,"prev=%p tmp=%p tmp->next=%p\n",prev,tmp,tmp->next);
    if (tmp->data == data) {
      flag=1;
      if (prev) prev->next=tmp->next;
      if (list==tmp) list=list->next;
      
      tmp->next=NULL;
      xfree(tmp);
      break;
    }
    prev=tmp;
    tmp=tmp->next;
  }
  
  if (flag==0) g_debug(0,"item not found to remove!!\n");
  g_debug(5,"new list=%p\n",list);
  return list;
}
  
// Gnut_List * gnut_list_prepend(Gnut_List *list, void *data)
//
// prepends data into the linked list list
// returns the new head pointer into the list
//
Gnut_List * gnut_list_prepend(Gnut_List *list, void *data)
{
  Gnut_List *gltmp;

  g_debug(5,"list=%p data=%p\n",list,data);
  
  gltmp=(Gnut_List *) xmalloc(sizeof(Gnut_List));
  assert(gltmp!=NULL);
  gltmp->data=data;
  gltmp->next=list;
  
  g_debug(5,"new list=%p\n",gltmp);
  return gltmp;
}

Gnut_List * gnut_list_append(Gnut_List *list, void *data)
{
  Gnut_List *gltmp;
  Gnut_List *glnew;

  glnew = (Gnut_List *) xmalloc(sizeof(Gnut_List));
  assert(glnew != NULL);
  glnew->data = data;
  glnew->next = NULL;

  if (list == NULL) {
    return glnew;
  } else {
    gltmp = list;
    while((gltmp->next) != NULL) {
      gltmp = gltmp->next;
    }
    gltmp->next = glnew;
    return list;
  }
}

// Gnut_List * gnut_list_next(Gnut_List *list)
//
// convenience function to return next item in list
//
Gnut_List * gnut_list_next(Gnut_List *list)
{
  return list->next;
}

// void gnut_list_foreach(Gnut_List * list, void (*GFunc)(void *, void *),
//       void *userdata)
//
// for each element in list, call GFunc
//
int gnut_list_foreach(Gnut_List *list, int (*GFunc)(void *,void *),
      void *userdata)
{
  Gnut_List *gltmp;
  
  g_debug(5,"list=%p\n",list);
  for (gltmp=list;gltmp!=NULL;gltmp=gltmp->next) {
    if ((*GFunc)(gltmp->data,userdata)==-1) return -1;
  }
  g_debug(5,"returning\n");
  return 0;
}

// Gnut_List * gnut_list_free(Gnut_List *list)
//
// free's all data associated with a list, including
// the data pointers; returns the new
// head pointer into the list
//
Gnut_List * gnut_list_free(Gnut_List *list)
{
  Gnut_List *gltmp,*gltmp2;
  
  g_debug(5,"list=%p\n",list);
  for (gltmp=list;gltmp!=NULL;gltmp=gltmp2) {
    g_debug(5,"gltmp=%p gltmp->next=%p\n",gltmp,gltmp->next);
    gltmp2=gltmp->next;
    xfree(gltmp->data);
    xfree(gltmp);
  }
  
  g_debug(5,"returning\n");
  return NULL;
}

// int gnut_list_size(Gnut_List *list)
//
// returns the number of items stored in list
//
int gnut_list_size(Gnut_List *list)
{
  Gnut_List *glt;
  int i;
  
  g_debug(5, "list=%p\n",list);
  
  for (i=0,glt=list;glt;glt=glt->next) {
    g_debug(5,"i=%i glt=%p glt->next=%p\n",i,glt,glt->next);
    i++;
  }
  
  g_debug(5, "returning: %i\n",i);
  
  return i;
}
 
