/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "gnut_lib.h"
#include "Gnut_Queue.h"

Gnut_Queue * gnut_queue_new()
{
  Gnut_Queue *a;
  
  a=(Gnut_Queue *) xmalloc(sizeof(Gnut_Queue));
  a->head=NULL;
  a->tail=NULL;
  a->size=0;
  return a;
}

void gnut_queue_delete(Gnut_Queue *q, void *d)
{
  Gnut_List *tmp;
  int flag=0;
  
  if (q->tail && d==q->tail->data) flag=1;
  
  q->head=gnut_list_remove(q->head,d);
  
  if (flag==1) {
    // we need to reset tail...
    for (tmp=q->head;tmp;tmp=tmp->next) {
      if (tmp->next==NULL) q->tail=tmp;
    }
  }
}  

void gnut_queue_insert(Gnut_Queue *q, void *d)
{
  Gnut_List *tmp;

  if (!q) {
    g_debug(0,"q=%p d=%p\n",q,d);
    exit(0);
  }
  g_debug(4,"entering q=%p d=%p\n",q,d);
  
  tmp=(Gnut_List *) xmalloc(sizeof(Gnut_List));
  tmp->data=d;
  tmp->next=NULL;
  q->size++;
  
  g_debug(4,"new size: %i\n",q->size);
  
  if (q->head==NULL) {
    q->head=q->tail=tmp;
  } else {
    gnut_mprobe(q->tail);
    q->tail->next=tmp;
    q->tail=tmp;
  }
}

void * gnut_queue_remove(Gnut_Queue *q)
{
  Gnut_List *gltmp;
  void *ret;
  
  gnut_mprobe(q);
  
  if (!q) {
    g_debug(0,"q=%p\n",q);
    exit(0);
  }
  
  gltmp=q->head;
  
  if (gltmp) {
    gnut_mprobe(gltmp);
    
    q->head=q->head->next;
    if (q->head==NULL) q->tail=NULL;
  
    ret=gltmp->data;
    free(gltmp);
  
    q->size--;
    return ret;
  } else {
    return NULL;
  }
}

Gnut_List * gnut_queue_list(Gnut_Queue *q)
{
  if (!q) {
    g_debug(0,"q=%p\n",q);
    exit(0);
  }
  return q->head;
}

void gnut_queue_free(Gnut_Queue *q)
{
  Gnut_List *gltmp,*gltmp2;
  
  g_debug(2,"entering q=%p\n",q);
  
  if (!q) {
    g_debug(0,"q=%p\n",q);
    exit(0);
  }
  
  for (gltmp=q->head;gltmp;gltmp=gltmp2) {
    gltmp2=gltmp->next;
    if (gltmp->data) free(gltmp->data);
    free(gltmp);
  }
  g_debug(2,"returning\n");
}
