/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "gnut_lib.h"
#include "gnut_threads.h"
#include "Gnut_Hash.h"
#include "conf.h"

conf_key_pair configuration_pairs[] = {
  {"ttl",     "4"},
  {"paginate",  "1"},
  {"local_port",  "5634"},
  {"max_results",  "64"},
  {"speed",    "28"},
  {"guid",    "AAAAAAAAAAAAAAAA"},
  {"query_guid",   "AAAAAAAAAAAAAAAA"},
  {"update_guid",  "AAAAAAAAAAAAAAAA"},
  {"min_connections","4"},
  {"stats_format", "0"},
  {"autokill_thres","10"},
  {"mac","ADCVFD"},
  {"auto_clear",  "1"},
  {"hidden",    "0"},
  {"share_paths",  ""},
  {"redirect",  "0"},
  {"download_path", ""},
  {"max_uploads", "10"},
  {"default_upload_cap", "0"},
  {"default_download_cap", "0"},
  {"download_timeout","3600"},
  {"max_incoming", "4"},
  {"global_bandwidth_cap", "0"},
  {"verbose","1"},
  {"play_prog","xmms"},
  {"cache_path",""},
  {"cache_size","15000"},  /* 15 megs.  Should this be measured in M instead of K? */
  {"cache_refresh","600"},
  {"packet_stats","0"},
  {"html_template",""},
  {"html_enable","1"},
  {"password",""},
  {"password_required","0"},
  {"auto_response","1"},
  {"max_responses","1000"},
  {"sort_order","s-p-"},
  {"update_clear","0"},
  {"beep_on_first","0"},
  {"beep_on_all","0"},
  {"munge","1"},
  {"max_downloads","5"},
  {"wait_after_find","1"},
  {(char *) NULL, (char *) NULL }
  };

Gnut_Hash *conf_hash=NULL;

int gc_ttl;

uchar conf_hash_struct(void *arg)
{
  conf_key_pair *ckp;
  int i,len;
  uchar ret;
  
  ckp=arg;
  len=strlen(ckp->key);
  for (i=0,ret=0;i<len;i++) ret+=ckp->key[i];
  
  return ret;
}

int conf_compare_struct(void *arg1, void *arg2)
{
  conf_key_pair *ckp1,*ckp2;
  
  ckp1=arg1;
  ckp2=arg2;
  
  return strcmp(ckp1->key,ckp2->key);
}

/* conf_cache_int saves certain configuration values which, for performance
 * reasons, must be kept in real global variables. It is called by each of
 * the other conf_routines, so the global is always in sync with the
 * configuration variable.
 *
 * Right now there's only one: the TTL */
void conf_cache_int(char *key)
{
  int val;

  if (strcmp(key, "ttl") == 0) {
  val = conf_get_int(key);
printf("set gc_ttl %i\n", val);
  gc_ttl = val;
  }
}

int conf_init()
{
  conf_key_pair *ckp;
  int i;
  
  conf_hash=gnut_hash_new(conf_hash_struct,conf_compare_struct);

  for(i=0;configuration_pairs[i].key!=NULL;i++) {
    ckp=(conf_key_pair *)xmalloc(sizeof(conf_key_pair));
    ckp->key=strdup(configuration_pairs[i].key);
    ckp->val=strdup(configuration_pairs[i].val);
    gnut_hash_insert(conf_hash,ckp);
    conf_cache_int(ckp->key);
  }
  
  return 0;
}

int conf_get_int(char *key)
{
  conf_key_pair ckp;
  conf_key_pair *ckp_ret;
  
  ckp.key=key;
  
  ckp_ret=gnut_hash_find(conf_hash,&ckp);
  
  if (!ckp_ret) {
    g_debug(0,"not found %s\n",key);
    return -1;
  }
  
  return atoi(ckp_ret->val);
}  

int conf_set_int(char *key, int val)
{
  conf_key_pair ckp;
    conf_key_pair *ckp_ret;

  ckp.key=key;
  
  ckp_ret=gnut_hash_find(conf_hash,&ckp);
  
  if (!ckp_ret) {
    g_debug(0,"not found: %s\n",key);
    return -1;
  }
  
  if (ckp_ret->val)
      xfree(ckp_ret->val);

  ckp_ret->val=xmalloc(20);
  sprintf(ckp_ret->val,"%i",val);

  conf_cache_int(key);

  return 0;
}

char *conf_get_str(char *key)
{
  conf_key_pair ckp,*ckp_ret;
  
  ckp.key=key;
  
  ckp_ret=gnut_hash_find(conf_hash,&ckp);
  
  if (!ckp_ret) {
    g_debug(0,"not found: %s\n",key);
    return NULL;
  }
  if (!ckp_ret->val) return NULL;
  
  return ckp_ret->val;
}

int conf_set_str(char *key, char *val)
{
  conf_key_pair ckp, *ckp_ret;
  
  ckp.key=key;
  
  ckp_ret=gnut_hash_find(conf_hash,&ckp);
  
  if (!ckp_ret) {
    g_debug(0,"not found: %s\n",key);
    return -1;
  }
  if (ckp_ret->val) xfree(ckp_ret->val);
  
  if (val) {
    ckp_ret->val=strdup(val);
  } else {
    ckp_ret->val=NULL;
  }

  conf_cache_int(key);
  
  return 0;
}

int conf_set_str_len(char *key, char *val, int len)
{
  conf_key_pair ckp, *ckp_ret;
    
  ckp.key=key;
  
  ckp_ret=gnut_hash_find(conf_hash,&ckp);
  
  if (!ckp_ret) {
    g_debug(0,"not found: %s\n",key);
    return -1;
  }
  
  if (ckp_ret->val)
    xfree(ckp_ret->val);
  
  if (val) {
    ckp_ret->val=(char *) xmalloc(len);
    memcpy(ckp_ret->val,val,len);
  } else {
    ckp_ret->val=NULL;
  }
  
  conf_cache_int(key);

  return 0;
}
