/* Josh Pieper, (c) 2000 */

#include <gnome.h>
#include <glade/glade.h>
#include <pthread.h>

#include "gnut_lib.h"
#include "Gnut_List.h"
#include "Gnut_Queue.h"
#include "gnut.h"
#include "gnut_net.h"
#include "monitor.h"
#include "share.h"
#include "gnut_connection.h"
#include "gnut_threads.h"
#include "host.h"
#include "protocol.h"
#include "conf.h"

int monitor_search_requests=1;
static int host_select_row=-1;

static GtkWidget *about_dlg;
gchar *authors[]={
  "Josh Pieper",
  "Kris Kennaway",
  "Damian M Gryski",
  NULL };

static GtkWidget *app;
static GtkWidget *host_list;
static uint  size_host_list=0;
static GtkWidget *catcher_list;
static GtkWidget *host_box;
static GtkWidget *connection_entry;
static GtkWidget *connections_text;
static GtkWidget *stats_text;
static GtkWidget *search_entry;
static GtkWidget *min_speed_entry;
static GtkWidget *search_response_list;

static GtkWidget *messages_info;
static GtkWidget *searches_info;
static GtkWidget *download_count_info;
static GtkWidget *upload_count_info;

int do_gnut_init()
{
  gnut_init();
  net_init(NULL);
  gnut_start_main_loop();
  return 0;
}

void on_Gnut_destroy(GtkWidget *a)
{
  exit(0);
}

void on_Search_clicked(GtkWidget *a)
{
  char *query;
  gnutella_packet *gpa;
  
  query=gtk_entry_get_text(GTK_ENTRY(search_entry));
  if (strlen(query)>1) {
//    gpa=gp_request_make(conf_get_mac(),query,conf_get_ttl());
    gpa=gp_request_make(conf_get_str("mac"),query,conf_get_int("ttl"));
//    conf_set_query_guid(gpa->gh.guid);
    conf_set_str_len("query_guid",gpa->gh.guid,16);
    query_clear();
    send_to_all(gpa);
    free(gpa->data);
    free(gpa);
  }
  
}

void on_Download_clicked(GtkWidget *a)
{
  printf("Download clicked\n");
}

void on_About_clicked(GtkWidget *a)
{
  about_dlg=gnome_about_new("gnut",
      VERSION,
      "(c) 2000 Josh Pieper",
      (const gchar **) authors,
      "Unix console/Gnome gnutella client.",
      (const gchar *) NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(about_dlg),GTK_WINDOW(app));
  gtk_widget_show(about_dlg);
}

void on_Update_clicked(GtkWidget *a)
{
  gnutella_packet *gpa;
  
//  gpa=gp_ping_make(conf_get_mac(),conf_get_ttl());
  gpa=gp_ping_make(conf_get_str("mac"),conf_get_int("ttl"));
  host_clear();
  
//  conf_set_update_guid(gpa->gh.guid);
  conf_set_str_len("update_guid",gpa->gh.guid,16);
  send_to_all(gpa);
  
  if (gpa->data) free(gpa->data);
  free(gpa);
}

void on_Remove_clicked(GtkWidget *a)
{
  gnut_connection *gc;
  gc=gtk_clist_get_row_data(GTK_CLIST(host_list),host_select_row);
  if (gc!=NULL) {
    gc->state=STATE_DEAD;
  }
}

void on_Add_clicked(GtkWidget *a)
{
  char *b;
  
  b=gtk_entry_get_text(GTK_ENTRY(host_box));
printf("new4\n");
  gnut_outgoing_new(b);
}


void on_host_list_select_row(GtkCList *clist,
  gint row, gint column, GdkEventButton *event, gpointer user_data)
{
  host_select_row=row;
}

void on_host_list_unselect_row(GtkCList *clist,
  gint row, gint column, GdkEventButton *event, gpointer user_data)
{
  host_select_row=-1;
}

int update_search_responses()
{
  static Gnut_List *lastpos=NULL;
  static uint32 lastsize=0;
  Gnut_List *tmp;
  Gnut_List *tmp2;
  char *list[4];
  query_response *qr=NULL;
  uint16 port;
  
  if (query_num()==lastsize) return 0;
  // we must have something different,
  // if cur size is less that last size, then
  // the list had to have been deleted at one point, which
  // means we need to clear out the list
  gtk_clist_freeze(GTK_CLIST(search_response_list));
  
  if (query_num()<lastsize) {
    gtk_clist_clear(GTK_CLIST(search_response_list));
    lastsize=0;
    lastpos=NULL;
  }
    
  // now we go through all the responses, making sure to
  // set lastpos and lastsize
  query_lock();
  tmp2=query_retrieve();
  for (tmp=tmp2;tmp!=lastpos;tmp=tmp->next) {
    lastsize++;
    qr=tmp->data;
    list[0]=qr->name;
    list[1]=xmalloc(20);
    sprintf(list[1],"%i",qr->speed);
    list[2]=xmalloc(40);
    memcpy(&port,qr->port,2);
    port=GUINT16_FROM_LE(port);
    sprintf(list[2],"%i.%i.%i.%i:%i",qr->ip[0],qr->ip[1],qr->ip[2],
      qr->ip[3],port);
    list[3]=xmalloc(10);
    sprintf(list[3],"%i",qr->ref);
    
    qr->index=gtk_clist_append(GTK_CLIST(search_response_list),list);
      
    free(list[1]);
    free(list[2]);
    free(list[3]);
  }
  query_unlock();
  gtk_clist_thaw(GTK_CLIST(search_response_list));
  lastpos=tmp2;    
  return 0;
}

int put_host_in_clist(gnut_connection *gc)
{
  char *list[3];
  int i;
  
  list[0]=(char *) xmalloc(100);
  if (gc->type==1) list[1]="Outgoing";
    else list[1]="Incoming";
  list[2]=(char *) xmalloc(100);
  sprintf(list[0],"%i.%i.%i.%i:%i",gc->ip[0],gc->ip[1],gc->ip[2],
    gc->ip[3],gc->port);
  sprintf(list[2],"%i/%i/%i",gc->sent_packets,
    gc->received_packets,
    gc->dropped_packets);

  i=gtk_clist_find_row_from_data(GTK_CLIST(host_list),gc);


  if (i==-1) { // we need to add the entry
    size_host_list++;
    i=gtk_clist_append(GTK_CLIST(host_list),list);
    gtk_clist_set_row_data(GTK_CLIST(host_list),i,gc);
  } else { // the entry is already in the table...
    gtk_clist_set_text(GTK_CLIST(host_list),i,2,list[2]);
  }
  
  free(list[0]);
  free(list[2]);
  return 0;    
}  

int update_host_list()
{
  int i,ret;
  gtk_clist_freeze(GTK_CLIST(host_list));
  
  gnut_connection_enumerate(put_host_in_clist);
  
  // now we'll go through and delete any rows that no longer have
  // connections associated with them
  for (i=0;i<size_host_list;i++) {
    ret=gnut_connection_find_ptr(gtk_clist_get_row_data(GTK_CLIST(host_list),i));
    if (ret==-1) { // we need to remove this row
      gtk_clist_remove(GTK_CLIST(host_list),i);
      i--;
      size_host_list--;
    }
  }
  
  gtk_clist_thaw(GTK_CLIST(host_list));    
  return 0;
}

int update_host_catcher()
{
//  struct ghost *gtemp;
//  char *list[3];
  
  gtk_clist_freeze(GTK_CLIST(catcher_list));
  gtk_clist_clear(GTK_CLIST(catcher_list));
  
//  printf("updating host catcher\n");
  
/*  start_enum_host();
  
  list[0]=(char *) xmalloc(100);
  list[1]=(char *) xmalloc(100);
  list[2]=(char *) xmalloc(100);
  
//  printf("entering loop\n");
  
  while ((gtemp=enum_host())) {
    sprintf(list[0],"%i.%i.%i.%i",gtemp->ip[0],gtemp->ip[1],
      gtemp->ip[2],gtemp->ip[3]);
    sprintf(list[1],"%i",gtemp->files);
    sprintf(list[2],"%i",gtemp->bytes/1024);
    
    gtk_clist_append(GTK_CLIST(catcher_list),list);
  }
  
  free(list[0]);
  free(list[1]);
  free(list[2]);
  
  end_enum_host();*/
  return 0;
}
  


int update_stats_boxes()
{
  char *a;
  int num;
  uint64 files,bytes;
  int messages,sent;
  uint64 rcv_bytes,sent_bytes;
  int queries,responses;

  a=(char *) xmalloc(200);
  sprintf(a,"%i gnutellaNet\n%i uploads\n%i downloads",
    gnut_connection_num(),gnut_threads_num_uploads(),
    gnut_threads_num_downloads());
  gtk_label_set_text(GTK_LABEL(connections_text),a);
  
  host_totals(&num,&files,&bytes);
  
  sprintf(a,"%i hosts\n%Li files\n%Li MB",
      num,files,bytes/1024);
  gtk_label_set_text(GTK_LABEL(stats_text),a);
  
  gnut_connection_net_totals(&messages,&sent,&rcv_bytes,&sent_bytes);
  
  sprintf(a,"%i",messages);
  gtk_label_set_text(GTK_LABEL(messages_info),a);
  
  gnut_connection_query_totals(&queries,&responses);
  
  sprintf(a,"%i",queries);
  gtk_label_set_text(GTK_LABEL(searches_info),a);

  sprintf(a,"%i",gnut_threads_count_uploads());
  gtk_label_set_text(GTK_LABEL(upload_count_info),a);
  
  sprintf(a,"%i",gnut_threads_count_downloads());
  gtk_label_set_text(GTK_LABEL(download_count_info),a);

  free(a);
  
//  if (num!=last_num) {
//    last_num=num;
//    update_host_catcher();
//  }
  
  return 0;
}

gint update_hosts(gpointer data)
{
  update_host_list();
  update_search_responses();
  update_stats_boxes();
  
  return 1;
}  

int main(int argc, char *argv[])
{
  GladeXML *xml;
  
  gnome_init("Gnut","0.3.0",argc,argv);
  
  glade_gnome_init();
  
  xml=glade_xml_new("gnut.glade","Gnut");
  
  if (!xml) {
    g_warning("Couldn't find my gnut.glade file!\n");
    return 1;
  }
  
  app=glade_xml_get_widget(xml,"Gnut");
  host_list=glade_xml_get_widget(xml,"host_list");
  catcher_list=glade_xml_get_widget(xml,"catcher_list");
  host_box=glade_xml_get_widget(xml,"host_entry");
  connection_entry=glade_xml_get_widget(xml,"connection_entry");
  connections_text=glade_xml_get_widget(xml,"connections_text");
  stats_text=glade_xml_get_widget(xml,"stats_text");
  messages_info=glade_xml_get_widget(xml,"messages_info");
  searches_info=glade_xml_get_widget(xml,"searches_info");
  download_count_info=glade_xml_get_widget(xml,"download_info");
  upload_count_info=glade_xml_get_widget(xml,"upload_info");
  search_entry=glade_xml_get_widget(xml,"search_entry");
  min_speed_entry=glade_xml_get_widget(xml,"min_speed_entry");
  search_response_list=glade_xml_get_widget(xml,"search_response_list");
  
  glade_xml_signal_autoconnect(xml);
  
  gtk_object_unref(GTK_OBJECT(xml));

  do_gnut_init();
  
  gtk_timeout_add(1000,update_hosts,NULL);
  
  gtk_main();
  
  return 0;
}

