/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/time.h>
#include <netdb.h>
#endif
#include <errno.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

#include "gnut_lib.h"
#include "gnut_threads.h"
#include "gnut_connection.h"
#include "Gnut_Queue.h"
#include "gnut_net.h"
#include "protocol.h"
#include "route.h"
#include "share.h"
#include "conf.h"
#include "host.h"
#include "monitor.h"
#include "query.h"
#include "cache.h"
#include "gnut.h"

#ifdef GNUT_HTTP_SEARCH
int gnut_http_result_append(
           uchar ip[4],
           uchar port[2],
           int ref,
           int speed,
           int size,
           char *name,
           uchar guid[16]
           );
#endif

// this mutex protects changes to the connection list
#ifndef PTHREADS_DRAFT4
pthread_mutex_t gc_list_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t send_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t query_packet_mutex = PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t gc_list_mutex;
pthread_mutex_t send_mutex;
pthread_mutex_t query_packet_mutex;
#endif

// linked list of all connections
Gnut_List *gc_list=NULL;

int num_queries=0;
int num_responses=0;

int num_messages=0;
int num_sent=0;

uint64 received_bytes=0;
uint64 sent_bytes=0;

int gnut_connection_num()
{
  return gnut_list_size(gc_list);
}

int gnut_connection_net_totals(int *messages, int *sent,
  uint64 *rcv_bytes, uint64 *sbytes)
{
  *messages=num_messages;
  *sent=num_sent;
  
  g_debug(2,"received_bytes=%li sent_bytes=%li\n",received_bytes,sent_bytes);
  *rcv_bytes=received_bytes;
  *sbytes=sent_bytes;
  return 0;
}

int gnut_connection_query_totals(int *queries,int *responses)
{
  *queries=num_queries;
  *responses=num_responses;
  return 0;
}

int gnut_connection_find_ptr(gnut_connection *gc)
{
  Gnut_List *gltmp;
  gnut_connection *mgc;

  for (gltmp=gc_list;gltmp;gltmp=gltmp->next) {
    mgc=gltmp->data;
    if (mgc==gc) {
      return 0;
    }
  }
  return -1;
}

/* int gnut_connection_find(uchar ip[4], uint16 port)
 *
 * Tests to see if the given IP/port is already on the active
 * connections list
 *
 * returns 0 if a connection is found to the listed ip and port
 * returns -1 otherwise
 */
int gnut_connection_find(uchar ip[4], uint16 port)
{
  Gnut_List *gltmp;
  gnut_connection *gc;

  for (gltmp=gc_list; gltmp; gltmp=gltmp->next) {
    gc=gltmp->data;
    if (memcmp(ip,gc->ip,4)==0 && gc->port==port) return 0;
  }

  return -1;
}

int send_to_one(gnut_connection *gc, gnutella_packet *gpa)
{
  gnutella_packet *ngp;
  
  g_debug(3,"entering (gc=%p, gpa=%p)\n",gc,gpa);
  
  pthread_mutex_lock(&gc->out_queue_mutex);

  g_debug(3, "done trying mutex\n");
  
  g_debug(3,"adding to packet queue\n");
  ngp=gp_dup(gpa);
  gnut_queue_insert(gc->out_queue,ngp);
  
  g_debug(3, "unlocking mutex\n");
  pthread_mutex_unlock(&gc->out_queue_mutex);

  if (gnut_lib_debug>=2) {
    g_debug(2,"I just sent this packet:\n");
    gp_print(gpa);
  }

  
  g_debug(3,"returning success\n");
  return 0;
}

int send_to_all(gnutella_packet *gpa)
{
  Gnut_List *gltemp;
  gnutella_packet *ngp;
  gnut_connection *gc;
  
  g_debug(3,"entering\n");

//  pthread_mutex_lock(&gc_list_mutex);
  
  for (gltemp=gc_list;gltemp;gltemp=gltemp->next) {
    gc=gltemp->data;
    
    if (gc->cstate==STATE_CONNECTED) {
      pthread_mutex_lock(&gc->out_queue_mutex);
    
      ngp=gp_dup(gpa);
      gnut_queue_insert(gc->out_queue,ngp);
    
      pthread_mutex_unlock(&gc->out_queue_mutex);
    }
  }
  
//  pthread_mutex_unlock(&gc_list_mutex);
  if (gnut_lib_debug>=2) {
    g_debug(2,"I just sent this packet:\n");
    gp_print(gpa);
  }
  
  g_debug(3,"returning success\n");
  return 0;
}
    
int send_to_all_except(gnut_connection *gc, gnutella_packet *gpa)
{
  Gnut_List *gltemp;
  gnutella_packet *ngp;
  gnut_connection *mgc;
  
  g_debug(3,"entering (gc=%p, gpa=%p)\n",gc,gpa);
  
//  pthread_mutex_lock(&gc_list_mutex);
  
  g_debug(3,"done locking\n");
  
  for (gltemp=gc_list;gltemp;gltemp=gltemp->next) {
    g_debug(3,"casting data\n");
    mgc=gltemp->data;
    g_debug(3,"checking for exceptions mgc=%p\n",mgc);
    if (mgc!=gc && mgc->cstate==STATE_CONNECTED) {
      g_debug(3,"adding to packet queue on mgc=%p\n",mgc);
      pthread_mutex_lock(&mgc->out_queue_mutex);
      g_debug(3,"mutex locked successfully\n");
      
      if (mgc->out_queue->size<BACKLOG_MAX) {
        g_debug(4,"duplicating packet\n");
        ngp=gp_dup(gpa);
        g_debug(4,"prepending to packet stack\n");
        gnut_queue_insert(mgc->out_queue,ngp);
      } else {
//        g_debug(1,"killing connection %p because of backlog\n",mgc);
//        mgc->cstate=STATE_DEAD;
      }
      
      pthread_mutex_unlock(&mgc->out_queue_mutex);
    }
  }

  if (gnut_lib_debug>=2) {
    g_debug(2,"I just sent this packet:\n");
    gp_print(gpa);
  }

  
//  pthread_mutex_unlock(&gc_list_mutex);
  g_debug(3,"returning success\n");
  return 0;
}

/* gnut_connection * gnut_connection_new()
 *
 * Allocates a new gnut_connection structure, fills in some default
 * values, and inserts it into the connection list, then returns
 * a pointer to this newly allocated structure.
 */
gnut_connection * gnut_connection_new()
{
  gnut_connection *gc;
  
  g_debug(3,"entering\n");
  
  gc=(gnut_connection *) calloc(sizeof(gnut_connection),1);
  
  gc->sock=-1;

#ifdef HP1020
#undef PTHREADS_DRAFT4
#endif
#ifndef PTHREADS_DRAFT4
  pthread_mutex_init(&gc->out_queue_mutex,NULL);
#else
  pthread_mutex_init(&gc->out_queue_mutex,pthread_mutexattr_default);
#endif

  gc->out_queue=gnut_queue_new();
  
  gc->sent_packets=0;
  gc->received_packets=0;
  gc->dropped_packets=0;
  gc->bad_packets=0;
  
  gc->sent_bytes=0;
  gc->received_bytes=0;
  
  gc->rate_sent_bytes=0;
  gc->rate_received_bytes=0;
  gc->rate_start_time=time(NULL);

  gc->routing_errors=0;
  
  gc->ctype=TYPE_UNSPEC;
  gc->cstate=STATE_UNCON;
  
  gc->tid=pthread_self();
  
  pthread_mutex_lock(&gc_list_mutex);
  
  gc_list=gnut_list_append(gc_list,gc); /* %%%RPM 0611: Add to end of list, to make connection numbers in 'info c' more stable */
  
  pthread_mutex_unlock(&gc_list_mutex);
  
  g_debug(3,"returning success\n");
  
  return gc;
}

// int gnut_connection_delete(gnut_connection *gc)
//
// takes the gnut_connection structure, free's up any memory
// associated with it, and removes it from the connection list
// returns 0 on success
//
int gnut_connection_delete(gnut_connection *gc)
{
  Gnut_List *gltmp;
  gnutella_packet *gpa;
  Gnut_List *unlinked;
  
  g_debug(1,"entering gc=%p\n",gc);
  
  pthread_mutex_lock(&gc_list_mutex);
  
  /* this didn't work, what we need to do is just unlink the list
   * then wait the second to let all references to it past,
   * then it will be safe to delete... */
  
  gc_list=gnut_list_unlink(gc_list,gc,&unlinked);
  
  pthread_mutex_unlock(&gc_list_mutex);
  
  g_debug(1,"unlinked\n");
  
  /* it's possible, although HIGHLY unlikely, that some other
   * function managed to get a hold on our pointer during out remove
   * operation... since I don't want to always lock the connection list
   * we need some other way of making sure that this particular pointer
   * isn't in use.  Perhaps a reference count of some sort....  sleeping
   * for a fraction of a second is probably an easier solution to
   * implement however */
  route_guid_clear(gc);
  
  sleep(2);
  
  free(unlinked);
  /* now that it's out of the connection list, we can
     go about freeing it up */
  
  for (gltmp=gc->out_queue->head;gltmp;gltmp=gltmp->next) {
    gpa=gltmp->data;
    if (gpa->data) free(gpa->data);
  }
  
  gnut_queue_free(gc->out_queue);

  pthread_mutex_destroy(&gc->out_queue_mutex); 

  free(gc);
  
  g_debug(1,"returning success\n");
  
  return 0;
}

// do whatever processing is necessary
int connection_handle_packet(gnut_connection *gc, gnutella_packet *gpa)
{
  gnut_connection *mgc;
  gnutella_packet *mgpa;
  gnutella_merchant *gm;
  gnutella_results *gr;
  gnutella_results_name *grn;
  gnutella_results_suffix *grs;
  gnutella_push *gp;
  uint32 num,size;
  host_entry *he;
  int isnew;
  Gnut_List *results;
  char *buf;
  uint32 speed;
  uint32 dlen;
  extern int monitor_search_requests;
  int i;
  
  isnew=0;
  
  g_debug(3,"entering gc=%p, gpa=%p\n",gc,gpa);

  memcpy(&dlen,gpa->gh.dlen,4);
  dlen=GUINT32_FROM_LE(dlen);
  
#ifdef MOREPACKETSTATS
  ++gc->packet_count[gpa->gh.func];
  if (dlen > gc->max_packet_size[gpa->gh.func])
    gc->max_packet_size[gpa->gh.func] = dlen;
  gc->packet_byte_count[gpa->gh.func] += dlen;
  ++gc->ttl_count[gpa->gh.ttl];
#endif

  if (dlen>65536) {
    g_debug(0, "WOAH! packet of size: %i made it through!\n",dlen);
    return -1;
  }

  if (gnut_lib_debug>=2) {
    g_debug(2,"received following packet\n");
    gp_print(gpa);
  }

  /* If the ttl is greater than 200, it's a packet that was forwarded by
   * a client with the "TTL wraparound bug", probably Windows Gnutella
   * prior to 0.56. These should be dropped. */
  if (gpa->gh.ttl>200 || gpa->gh.hops>200) {
    g_debug(2,"ttl too high: ttl=%i\n",gpa->gh.ttl);
  /* printf("."); %%%RPM */
    return 0;    
  }

  /* Any other packets with a TTL bigger than our TTL should have their
   * TTL diminished, and if their hopcount is greater than our TTL we
   * should drop 'em */
  if (gpa->gh.ttl > gc_ttl) {
    gpa->gh.ttl = gc_ttl;
  /* printf("-"); %%%RPM */
  }
  if (gpa->gh.hops > gc_ttl) {
    g_debug(2,"hops too high: %i\n",gpa->gh.hops);
  if (0) {
    /* %%%RPM diagnostics: bad TTLs implies strange data in packets,
     * perhaps ASCII data like query replies or chatroom messages */
    int j; char c; char *t;
    printf("%i'", gpa->gh.hops);
    t = (char *) (&(gpa->gh.guid));
    for(j=0; j<22; j++) {
    c = *t++;
    if (c>31 && c<127)
      printf("%c", c);
    }
    printf("'");
  }
    return 0;
  }

  memcpy(&dlen, gpa->gh.dlen, 4);
  dlen=GUINT32_FROM_LE(dlen);

  /* Check for GUID's we've seen before, and locate the connection
   * that GUID originated from. If it's a known GUID we might drop the
   * packet, or we might route it to that particular connection, depending
   * on the type of packet. */
  mgc=route_guid_find(gpa->gh.guid);
  if (mgc)
    isnew=0;
  else
    isnew=1;

  g_debug(3,"entering packet switch\n");
    
  switch(gpa->gh.func) {
  case 0:
    if (isnew==1 && !conf_get_int("hidden")) {
      g_debug(3,"ping packet\n");
      if (dlen>0) {
        g_debug(1,"dlen=%i\n",dlen);
        return -2;
      }
      g_debug(2,"handling ping packet\n");
      // this is a ping packet, we should reply
      // to it with a merchant packet...
      share_totals(&num, &size);
      
      mgpa=gp_merchant_make(gpa->gh.guid, num, size, 
        net_local_ip() , conf_get_int("local_port") , 
        conf_get_int("ttl"));
      
      send_to_one(gc, mgpa);
    
      /* We're done with the packet; free it. */
      if (mgpa->data)
        free(mgpa->data);
      free(mgpa);
    } else if (isnew == 0) {
      gc->dropped_packets++;
    }
    break;

  case 1:
    if (dlen!=sizeof(gnutella_merchant)) {
      g_debug(1,"bad merchant packet dlen=%i should be=%i\n",
        dlen,sizeof(gnutella_merchant));
      return -3;
    }
    gm=gpa->data;
//    if (gm->ip[0]==127 || gm->ip[0]==10) {
//      // unrouteable IP, this packet does no
//      // good, drop it
//      g_debug(2,"dropping unrouteable ip\n");
//      return 0;
//    }

  g_debug(2,"handling merchant packet\n");
  if (host_find(gm->ip,gm->port) == -1) {
    if (host_ok_for_connect(gm->ip)) {
        g_debug(2,"adding new merchant\n");
        /* the host wasn't found in the database, so we'll add it */
        memcpy(&num,gm->files,4);
        num=GUINT32_FROM_LE(num);
        
        memcpy(&size,gm->bytes,4);
        size=GUINT32_FROM_LE(size);
      
        he=(host_entry *) xmalloc(sizeof(host_entry));
        memcpy(he->ip,gm->ip,4);
        memcpy(he->port,gm->port,2);
        
        he->files=num;
        he->bytes=size;
        host_add(he);
    }
  }
    if (memcmp(conf_get_str("update_guid"),gpa->gh.guid,16)==0) {
      return 0;
    }
//      g_debug(3,"not our merchant packet\n");
//    }
    break;

  case 0x40:
    if (dlen!=sizeof(gnutella_push)) {
      g_debug(1,"bad push request! dlen=%i should be=%i\n",
        dlen,sizeof(gnutella_push));
      return -4;
    }
//    if (isnew==1) {

      gp=gpa->data;
      g_debug(3,"push request\n");
      if (memcmp(gp->guid,conf_get_str("guid"),16)==0) {
        g_debug(2,"my own push request\n");
        gnut_push_new(gp);
        // return because we don't want to route this sucker
        return 0;    
      }
      // now we're going to see if this is in our routing table, if
      // it is, then we send to just that one and quit, otherwise we will
      // let it drop through into the regular routing code which will end
      // up broadcasting it
      
      // screw that, only do pushes that we saw the original query for
      // that should help congestion a bit
      mgc=route_guid_find(gp->guid);
      if (mgc) {
        g_debug(2,"routing push request!\n");
        // here we do the routing
        gpa->gh.hops++;
        if (gpa->gh.ttl!=0 && --gpa->gh.ttl>0) {
          send_to_one(mgc,gpa);
        } else {
          /* gc->dropped_packets++; */
        }
        
        return 0;
      } else return 0;
//    } else return 0;

    break;

  case 0x80:
    if (dlen==0) return -5;
    g_debug(3,"query request\n");
    // we need to run the query!!!!
    if (isnew) {
      num_queries++;
      g_debug(3,"new query request\n");
      buf=gpa->data;
      buf[dlen-1]=0;
      results=share_search(&buf[2],conf_get_int("max_results"));
      
      if (monitor_search_requests) {
        monitor_search_add(&buf[2], !!results);
      }
      
      if (results) {
        num_responses++;
        g_debug(3,"responding to query: %s\n",&buf[2]);
        
        mgpa=gp_reply_make(gpa->gh.guid,results,
          net_local_ip(), conf_get_int("local_port"), conf_get_int("speed"),
          conf_get_str("guid"), conf_get_int("ttl"));
        
        send_to_one(gc,mgpa);
        
        share_clear_list(results);
        
        free(mgpa->data);
        free(mgpa);
      }
    } else if (isnew == 0) {
      gc->dropped_packets++;
    }
    break;

  case 0x81:
    if (dlen<sizeof(gnutella_results)) {
      g_debug(1,"dlen=%i should be>%i\n",dlen,sizeof(gnutella_results));
      return -7;
    }
    g_debug(3,"handling query_reply\n");

    grs=(gnutella_results_suffix *) ((char *) gpa->data + dlen - 16);
    /* we stick this into the routing table, so that
     * it is possible to route push connections back to the
     * place that originally sent us the query... */
    route_guid_add(grs->guid,gc);

#ifdef GNUT_HTTP_SEARCH
    {
    int pref,psize;
    int ret;

    gr=gpa->data;
    grn=(gnutella_results_name *) ((char *) gpa->data+sizeof(gnutella_results));
    grs=(gnutella_results_suffix *) ((char *) gpa->data + dlen - 16);

    memcpy(&pref,grn->ref,4);
    pref=GUINT32_FROM_LE(pref);
        
    memcpy(&psize,grn->size,4);
    psize=GUINT32_FROM_LE(psize);

    memcpy(&speed,gr->speed,4);
    speed=GUINT32_FROM_LE(speed);
      
    ret = gnut_http_result_append(
          gr->ip,
          gr->port,
          pref,
          speed,
          psize,
          grn->name,
          gpa->gh.guid
          );
    /*
     * If this was requested through
     * an HTTP request, append it to the
     * appropriate cache, and not to our interactive
     * query list.
     */
    if (ret) break;
    }
#endif

  /* We now only compare the first 14 bytes of the query GUID, because
   * the last two are used for distinguishing results from multiple searches
   * (not yet implemented) */
    if (memcmp(conf_get_str("query_guid"), gpa->gh.guid, 14)==0) {
    uchar the_guid[16];

    memcpy(the_guid, gpa->gh.guid, 14);
    the_guid[14] = 0;
    the_guid[15] = 0;

      g_debug(3,"its our query reply\n");
      gr=gpa->data;
            
      memcpy(&speed,gr->speed,4);
      speed=GUINT32_FROM_LE(speed);
            
      grn=(gnutella_results_name *) ((char *) gpa->data+sizeof(gnutella_results));
      for (i=0;i<gr->num;i++)  {
        int pref, psize;
        
        if (((char *)&grn->name - (char *)gpa->data) > dlen) {
          g_debug(1,"bad response packet\n");
          return -8;
        }
        
        memcpy(&pref, grn->ref, 4);
        pref=GUINT32_FROM_LE(pref);
        
        memcpy(&psize, grn->size, 4);
        psize=GUINT32_FROM_LE(psize);

        query_add(gr->ip, gr->port, pref, speed, psize, grn->name, the_guid);
        
        grn=(gnutella_results_name *) ((char *) grn+sizeof(gnutella_results_name) + strlen(grn->name)+1);
      }
      return 0;
    }
  
  // if we did not search for this query response, possibly cache it
  if (cache_time_for_update()) {
    int pref,psize;
    g_debug(1,"caching from query reply\n");
    gr=gpa->data;
    
      memcpy(&speed,gr->speed,4);
      speed=GUINT32_FROM_LE(speed);
    
      // no loop... just cache the first in the replies (FIXME - Ray)
      grn=(gnutella_results_name *) ((char *) gpa->data+sizeof(gnutella_results));
      if (((char *)&grn->name - (char *)gpa->data) > dlen) {
        g_debug(1,"bad response packet\n");
        return -8;
      }
    
      memcpy(&pref,grn->ref,4);
      pref=GUINT32_FROM_LE(pref);
    
      memcpy(&psize,grn->size,4);
      psize=GUINT32_FROM_LE(psize);
    
      query_add_cache(gr->ip,gr->port,pref,speed,psize,grn->name,grs->guid);
    
      return 0;
    }

    break;  

  default:
    // return here so we don't pass along bad packets...
    g_debug(1, "packet type: %i\n", gpa->gh.func);
    return -9;
  }

  g_debug(2,"running routing logic\n");

  if ((gpa->gh.func & 1)!=0) {
    if (mgc==NULL) {
      g_debug(2,"Routing error!\n");
      gc->routing_errors++;
      return 0;
    } else {
      g_debug(2,"Routing packet...\n");
      gpa->gh.hops++;
      if (gpa->gh.ttl!=0 && --gpa->gh.ttl>0) {
        send_to_one(mgc, gpa);
      } else {
        /* gc->dropped_packets++; */
      }
    }
  } else {
    if (!mgc) {
      g_debug(2,"Spreading packet!\n");
    
      route_guid_add(gpa->gh.guid,gc);
      
      gpa->gh.hops++;
      if (gpa->gh.ttl!=0 && (--gpa->gh.ttl)>0) {
        send_to_all_except(gc, gpa);
      } else {
        /* gc->dropped_packets++; */
      }
    }
  }
  
  g_debug(3,"returning success\n");
  return 0;
}

/* loops through any packets available on the out stack,
 * and sends them in turn...
 */
int connection_send_packets(gnut_connection *gc)
{
  gnutella_packet *gpa;
  int ret;
  fd_set fsw;
  int i=0;
  struct timeval tv;
  int sendit;

  g_debug(4,"entering\n");

  while (gc->out_queue->size>0) {
    g_debug(3,"there are %i packets left to send\n",
      gc->out_queue->size);

    FD_ZERO(&fsw);
    FD_SET(gc->sock,&fsw);

  sendit = 1;

  /* Wait 0.5 seconds for socket to be writable. What this really means
       is we're waiting for the current I/O to complete. It takes longer
       if the connection at the other end is slow */
    tv.tv_sec=1;
    tv.tv_usec=500000;
    ret=select(gc->sock+1, NULL, &fsw, NULL, &tv);

    if (ret==0) {
    /* printf("."); %%%RPM */
      sendit = 0;
    }

  /* we don't want anyone trying to change the out stack
   * while we're playing around with it...
   */
  pthread_mutex_lock(&gc->out_queue_mutex);
  gpa=gnut_queue_remove(gc->out_queue);
  pthread_mutex_unlock(&gc->out_queue_mutex);

    if (sendit) {
      ret=send_packet(gc->sock, gpa);

    if (ret>=0) {
    g_debug(4,"wrote packet, size: %i\n",ret);
    gc->sent_bytes+=ret;
    gc->rate_sent_bytes+=ret;
    gc->sent_packets++;
    sent_bytes+=ret;
    num_sent++;
    }
    
#ifdef MOREPACKETSTATS
    ++gc->out_packet_count[gpa->gh.func];
    gc->out_packet_byte_count[gpa->gh.func] += (ret - 23);
    ++gc->out_ttl_count[gpa->gh.ttl];
    if ((ret - 23) > gc->out_max_packet_size[gpa->gh.func])
    gc->out_max_packet_size[gpa->gh.func] = (ret - 23);
#endif
    } else {
    /* We're gonna just drop the packet. Better than dropping the whole connection! */
      gc->dropped_packets++;
      ret = 0;
    }

    if (gpa->data)
      free(gpa->data);
    free(gpa);

    if  (ret<0) {
    /* printf("csp send_packet error %s\n", strerror(errno));  %%%RPM */
      g_debug(1, "csp send_packet error %s", strerror(errno));
      return ret;
    }
    
    /* Limit to sending 5 packets at a time. This is so we can still receive stuff over the link... */
    if ((++i)>5)
      return 0;
  }

  g_debug(4,"returning success\n");
  
  return 0;
}

// this needs to read a packet if available, and it NEEDS to timeout
// after 1/4 second or less
// returns 0 on success or timeout
// returns <0 on error
int connection_receive_packet(gnut_connection *gc)
{
  fd_set fsr;
  struct timeval tv;
  int ret;
  gnutella_packet * gpa;
  
  g_debug(4,"entering\n");
  
  FD_ZERO(&fsr);
  FD_SET(gc->sock,&fsr);
  
  tv.tv_sec=0;
  tv.tv_usec=250000;
  ret=select(gc->sock+1, &fsr, NULL, NULL, &tv);
  
  if (ret<=0) {
//    perror("connection_receive_packet, select");
    return ret;
  }
  
  g_debug(3,"trying to read packet\n");
  gpa=(gnutella_packet *) calloc(sizeof(gnutella_packet),1);
  
  gnut_mprobe(gpa);
  ret=read_packet(gc->sock,gpa);
  
  gnut_mprobe(gpa);
  
  if (ret<0 && errno!=EINTR) {
    g_debug(1,"read_packet returned:%i\n",ret);
    if (gpa->data) free(gpa->data);
    free(gpa);
    return ret;
  }
  
  if (ret>0) {
    received_bytes+=(uint64) ret;
    gc->received_bytes+=(uint64) ret;
    gc->rate_received_bytes+=(uint64) ret;
    gc->received_packets++;
    num_messages++;
  
    // now we can do any actual processing on the packet!
  
    ret=connection_handle_packet(gc,gpa);
  
    gnut_mprobe(gc);
    gnut_mprobe(gpa);
    if (gpa->data) gnut_mprobe(gpa->data);
  
    if (ret<0) {
      g_debug(1,"bad packet on tid: %li with ret=%i\n",gc->tid,ret);
      gc->bad_packets++;
      if (gc->bad_packets>BAD_PACKET_MAX) {
        g_debug(1,"max bad packets reached on this connection\n");
        return -1;
      }
      g_debug(2,"ret=%i\n",ret);
    }
  
    g_debug(3,"trying to free gpa->data\n");
  
    if (gpa->data) {
      gnut_mprobe(gpa->data);
      free(gpa->data);
    }
  }
  g_debug(3,"trying to free gpa\n");
  
  gnut_mprobe(gpa);
  free(gpa);
  
  g_debug(3,"returning success\n");
  
  return 0;
}


/* int connection_loop(gnut_connection *gc)
 *
 * this is the main loop which handles sending and receiving
 * all packets over a connection.  It should be called after
 * all initial negotiation is complete.  When it returns, it should
 * be time to remove the connection from the connection
 * list.
 */
int connection_loop(gnut_connection *gc)
{
  int ret=1;

  g_debug(3,"entering\n");

  gc->cstate=STATE_CONNECTED;

  /* If there is a current query, we send it out the new connection.
   * The effect of this is that, if you issue a 'find' command right
   * after starting gnut, you'll get just as many results as you would
   * have gotten if you waited for all 4 (or however many) connections
   * to be up before doing the 'find'. This is good, because it means
   * you don't have to wait every time you launch gnut, or (even
   * worse :-) repeat the search command 'just to make sure'. */
  if (query_num() < conf_get_int("max_responses")) {
  pthread_mutex_lock(&query_packet_mutex);
  if (current_query_packet) {
    /* There is a query (find) in progress, and we have just established
     * a new connection into the network.
     *
     * Because this new connection gives us access to hosts that were
     * hitherto unreachable within the TTL, we want to send out the
     * query packet so those new hosts can get the query and respond to it.
     *
     * However, we also have to change the GUID in the query. If we don't,
     * and if the host we've just connected to has already seen the query
     * (which is very likely, since we got the host's IP address from a
     * packet and therefore the host was within TTL of a previous connection)
     * it would just throw the packet away and the packet would not reach our
     * new hosts. Also, if a match occurred, the results would get routed
     * through the old path, but we want them to come back on the new path. */
    if (1) { /* %%%RPM disabled, because it causes lots of duplicate results */
    current_query_packet->gh.guid[14]++;
    if (current_query_packet->gh.guid[14] == 0) {
      current_query_packet->gh.guid[15]++;
    }
    }
    send_to_one(gc, current_query_packet);
  }
  pthread_mutex_unlock(&query_packet_mutex);
  }

  while((gc->cstate==STATE_CONNECTED) && (ret>=0)) {
    ret = connection_receive_packet(gc);
    g_debug(3, "connection_receive_packet returned: %i\n", ret);
    if (ret<0) {
      // the line is closed, we must be done!
      g_debug(1,"connection_receive_packet returned: %i\n",ret);
    /* printf("%d.%d.%d.%d: close1\n", gc->ip[0], gc->ip[1], gc->ip[2], gc->ip[3]); %%%RPM */
    }

    if (ret>=0) {
      /* send all queued up packets */
      g_debug(3,"about to call connection_send_packets with gc=%p\n",gc);
      ret=connection_send_packets(gc);
      g_debug(3,"send_packets returned: %i\n",ret);
      if (gc->cstate==STATE_CONNECTED)
        g_debug(20,"this shouldn't get written\n");
      g_debug(3,"gc->cstate is readable\n");
      if (ret<0) {
    /* printf("%d.%d.%d.%d: close3\n", gc->ip[0], gc->ip[1], gc->ip[2], gc->ip[3]); %%%RPM */
      }
    }
  }

  if (ret >= 0) {
    /* We exited on cstate */
  /* printf("%d.%d.%d.%d: close2\n", gc->ip[0], gc->ip[1], gc->ip[2], gc->ip[3]); %%%RPM */
  }

  g_debug(3,"returning success\n");
  return 0;
}

// calls function a for every connection in the list
int gnut_connection_enumerate( int (*a)(gnut_connection *))
{
  Gnut_List *gl;
  
  pthread_mutex_lock(&gc_list_mutex);
  
  for (gl=gc_list;gl;gl=gl->next) {
    if ((*a)(gl->data)==-1) break;
  }
  
  pthread_mutex_unlock(&gc_list_mutex);
  return 0;
}

// kill the oldest incoming connection
int gnut_connection_kill_oldest()
{
  Gnut_List *gl;
  gnut_connection *gc=NULL;
  gnut_connection *gtc;
  
  pthread_mutex_lock(&gc_list_mutex);
  
  for (gl=gc_list;gl;gl=gl->next) {
    gtc=gl->data;
    if ((gtc->ctype==TYPE_INCOMING) && (gtc->cstate==STATE_CONNECTED)) {
      gc=gtc;
    }
  }
  if (gc) {
    // the instance of connection_loop() handling this connection
    // will notice this and actually close the connection
    gc->cstate=STATE_DEAD;
  }

  pthread_mutex_unlock(&gc_list_mutex);
  return 0;
}

/* droprate returns ratio of two longs, used to compute
   ratio of dropped to total received packets. Result is an
   8-bit fractional number: 256 means 1.0 or above; 0 means 0.0;
   in-between is a fraction, for example 192 means 0.75 */
long droprate(long drop, long recv)
{
  long rv;

  if ((recv == 0) || (drop == 0)) {
    rv = 0;
  } else if (drop > recv) {
    rv = 256;
  } else if (drop > 65536L) {
    rv = drop / (recv >> 8L);
  } else {
    rv = (drop << 8L) / recv;
  }

  // convert to a percentage
  rv = ((rv * 100) + 128) / 256;

  return rv;
}

// kill the worst (by dropped packet percentage) connection
int gnut_connection_kill_worst(int interact)
{
  Gnut_List *gl;
  gnut_connection *worst_gc=NULL;
  gnut_connection *gtc;
  long this_dr;
  long worst_dr;
  int killed_ip[4];

  pthread_mutex_lock(&gc_list_mutex);

  worst_dr = conf_get_int("autokill_thres");
  for (gl=gc_list;gl;gl=gl->next) {
    gtc=gl->data;
    if (((gtc->ctype==TYPE_OUTGOING) || (gtc->ctype==TYPE_INCOMING))
         && (gtc->cstate==STATE_CONNECTED)
         && ((gtc->sent_packets + gtc->received_packets + gtc->dropped_packets) > 1000))
    /* %%%RPM also want to test for really low transfer/sec stats. It needs to measured relative to other connections, to auto-compensate for our ISP speed */
    {
      this_dr = droprate(gtc->dropped_packets, gtc->received_packets);

    /* 20000611 change > to >=; break ties by last found rather than first; (this is newest connection after reversing order of connection list insertion) */
      if (this_dr >= worst_dr) {
        worst_dr = this_dr;
        worst_gc=gtc;
      }
    }
  }
  if (worst_gc) {
    // record the IP of the connection we killed (no printf inside mutex)
    killed_ip[0] = worst_gc->ip[0];
    killed_ip[1] = worst_gc->ip[1];
    killed_ip[2] = worst_gc->ip[2];
    killed_ip[3] = worst_gc->ip[3];

    // set flag to kill connection:
    // the instance of connection_loop() handling this connection
    // will notice this and actually close the connection
    worst_gc->cstate=STATE_DEAD;
  }

  pthread_mutex_unlock(&gc_list_mutex);

  if (interact) {
    /* indicate which , if any, we killed */
    if (worst_gc) {
      printf("Killing connection to %d.%d.%d.%d\n", killed_ip[0],
        killed_ip[1], killed_ip[2], killed_ip[3]);
    } else {
      printf("No worst connection found.\n");
    }
  }

  return 0;
}
