/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#ifdef MCHECK
#include <mcheck.h>
#endif

#include "gnut_lib.h"
#include "gnut_threads.h"

int _gnut_mprobe(void *d)
{
#ifdef MCHECK
  return mprobe(d);
#else
  return 0;
#endif
}

void _xfree(void **a)
{
  if (*a==NULL) {
    g_debug(0,"trying to free NULL ptr\n");
    return;
  }
  free(*a);
  *a=NULL;
}

#ifndef HAVE_XMALLOC
#ifndef xmalloc
char *xmalloc(int size)
{
  char *ptr;
  ptr=malloc(size);
  if (!ptr) { g_debug(0,"malloc(%d) failed!\n", size); exit(1); }
  return ptr;
}
#endif
#endif

#ifdef WIN32
char *path_separator = ";";
char *path_slash = "\\";
#else
char *path_separator = ":";
char *path_slash = "/";
#endif

char *expand_path(char *a)
{
  char *b;
  if (strncmp(a,"~/", 2)==0) {
#ifndef WIN32
    char *c;
    c=getenv("HOME");
#else
    char c[MAX_PATH];
//    GetWindowsDirectory(c, MAX_PATH);
    // I did this because too many people couldn't find
    // the .gnut* files under win32  -jp
    c[0]=0;  
#endif
    b=(char *)xmalloc(strlen(a)+strlen(c)+3);
    b[0]=0;
    if (strlen(c)>1) {
      strcpy(b, c);
      strcat(b, path_slash);
    }
    strcat(b, &a[2]);
  } else {
    b=(char *) xmalloc(strlen(a)+3);
    strcpy(b,a);
  }
  
  
  if (b[strlen(b)-1]==path_slash[0]) b[strlen(b)-1]=0;
  
  return b;   
}                                               

int gnut_lib_debug=0;

#ifndef PTHREADS_DRAFT4
pthread_mutex_t _g_debug_mutex=PTHREAD_MUTEX_INITIALIZER;
#else
pthread_mutex_t _g_debug_mutex;
#endif

#ifndef WIN32
void _g_debug(char *file, int line, int num, char *format, ...)
{
  va_list argp; 
       
  if (gnut_lib_debug >= num) {
    pthread_mutex_lock(&_g_debug_mutex);
    fprintf(stderr,"%li %s:%i > ",
      pthread_self(),file,line);

    fflush(stdout);
    va_start(argp,format);
    vfprintf(stderr,format,argp);
    va_end(argp);
    pthread_mutex_unlock(&_g_debug_mutex);
    }
}
#else
void g_debug(int a, char *format, ...)
{
  va_list argp;
  if (gnut_lib_debug>=a) {
    pthread_mutex_lock(&_g_debug_mutex);
    fprintf(stderr,"%li > ", (long) pthread_self());
    va_start(argp,format);
    vfprintf(stderr,format,argp);
    va_end(argp);
    pthread_mutex_unlock(&_g_debug_mutex);
  }
}
#endif

char *gnut_strdelimit(char *string, char *delim, char new_delim)
{
  char *c;
  
  for (c=string;*c;c++) {
    if (strchr(delim,*c)) *c=new_delim;
  }
  
  return string;
}

// remove any trailing whitespace
void gnut_strstrip(char *string)
{
  int i;
  for (i=(strlen(string)-1);i>0 && isspace(string[i]);i--) string[i]=0;
}

char *munge_time(int secs)
{
  /* these buffers should be 'large enough' */
  char *s;
  char buf[10];
  int days,hours,minutes,seconds;
  
  s=xmalloc(40*sizeof(char));
  days=hours=minutes=seconds=0;
  strcpy(s,"");
  
  if (secs==0) {
    strcat(s,"0s");
    return s;
  }
  
  if (secs<0) {
    /* handle negatives correctly */
    strcat(s,"-");
    secs=-secs;
  }
  
  seconds=secs;
  
  if (seconds > 60) {
    /* how many minutes? */
    minutes=seconds / 60;
    seconds=seconds % 60;
  }
  
  if (minutes > 60) {
    /* hours? */
    hours=minutes / 60;
    minutes = minutes %60;
  }
  
  if (hours > 24) {
    /* and the days */
    days=hours/24;
    hours=hours % 24;
  }
  
  /* build the string */
  if (days) { sprintf(buf,"%dd",days); strcat(s,buf); }
  if (hours) { sprintf(buf,"%dh",hours); strcat(s,buf); }
  if (minutes) { sprintf(buf,"%dm",minutes); strcat(s,buf); }
  if (seconds) { sprintf(buf,"%ds",seconds); strcat(s,buf); }
  
  return s;
}


char *munge_bytes(uint64 bytes) {
  /* these buffers should be 'large enough' */
  char *s;
  
  s=xmalloc(50*sizeof(char));
  
  if (bytes<1024) {
    sprintf(s, "%Li", bytes);
    return s;
  }
  
  /* here are some defines to make the following code cleaner */
  /* note that somewhere along the line, the defn of an MB changed */
  /* these could probably be moved to a .h somehwere */
  
#define MEGABYTE (uint64)(1024 * 1024)
#define GIGABYTE (uint64)(1000 * MEGABYTE)
#define TERABYTE (uint64)(1000 * GIGABYTE)

  if ((bytes >= 1024) && (bytes < MEGABYTE)) {
    /* we only need one digit after the decimal for kilobytes */
    sprintf(s, "%.2fK", ((float)bytes/1024.0));
    return s;
  }
  
  if ((bytes >= MEGABYTE) && (bytes < GIGABYTE)) {
    sprintf(s,"%.3fM", ((float)bytes/MEGABYTE));
    return s;
  }
  
  if ((bytes >= GIGABYTE) && (bytes < TERABYTE)) {
    sprintf(s,"%.3fG", ((float)bytes/GIGABYTE));
    return s;
  }
  
  sprintf(s,"%.3fT",((float)bytes/TERABYTE));
  return s;
  
#undef MEGABYTE
#undef GIGABYTE
#undef TERABYTE
}
                       
