/* Jon Arney, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#endif

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <assert.h>

#include "gnut_lib.h"
#include "gnut_connection.h"
#include "gnut_net.h"
#include "host.h"
#include "gnut_transfer.h"
#include "query.h"
#include "conf.h"
#include "share.h"
#include "gnut_threads.h"
#include "gnut_http.h"
#include "Gnut_Hash.h"

/*
 * Create a new header object.
 */
http_header_t *http_header_new(void)
{
  http_header_t *hdr;

  g_debug(HTTP_HEADER_DEBUG, "IN\n");

  hdr = (http_header_t *)xmalloc(sizeof(http_header_t));
  assert(hdr!=NULL);
  hdr->name = NULL;
  hdr->value = NULL;

  g_debug(HTTP_HEADER_DEBUG, "OUT\n");

  return hdr;
}

/*
 * Parse an HTTP header from a header line.
 */
int http_header_parse(http_header_t *hdr, char *hdr_line)
{
  char *value;
  char *s;
  int len;

  g_debug(HTTP_HEADER_DEBUG, "IN\n");

  value = strchr(hdr_line, ':');
  if (!value) {
    g_debug(HTTP_HEADER_DEBUG, "This is not a proper header\n");
    return -1;
  }

  len = value - hdr_line;
  hdr->name = xmalloc(sizeof(char)*(len+1));
  assert(hdr->name!=NULL);

  strncpy(hdr->name, hdr_line, len);
  *(hdr->name+len) = '\0';

  /*
   * Skip past the ":" and the space.
   */
  value++;
  if (*value) value++;

  if (!(*value)) {
    xfree(hdr->name);
    xfree(hdr);
    g_debug(HTTP_HEADER_DEBUG, "OUT - not a proper header\n");
    return -1;
  }

  len = strlen(value)-2;
  hdr->value = xmalloc(sizeof(char)*(len+1));
  assert(hdr->value!=NULL);

  strncpy(hdr->value, value, len);
  *(hdr->value+len) = '\0';

  s = hdr->value+(len-1);
  while (isspace(*s)) *s = '\0';

  g_debug(HTTP_HEADER_DEBUG, "OUT\n");

  return 0;
}

/*
 * Set the contents of an HTTP header.
 */
int http_header_set(http_header_t *hdr, char *name, char *value)
{
  g_debug(HTTP_HEADER_DEBUG, "IN\n");
  hdr->name = strdup(name);
  hdr->value = strdup(value);
  g_debug(HTTP_HEADER_DEBUG, "OUT\n");
  return 0;
}

void http_header_delete(http_header_t *hdr)
{
  g_debug(HTTP_HEADER_DEBUG, "IN\n");
  if (hdr) {
    if (hdr->name) xfree(hdr->name);
    if (hdr->value) xfree(hdr->value);
    xfree(hdr);
  }
  g_debug(HTTP_HEADER_DEBUG, "OUT\n");
}

