/* Josh Pieper, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "gnut_lib.h"
#include "share.h"
#include "Gnut_List.h"
#include "Gnut_Queue.h"
#include "Gnut_Hash.h"
#include "share_cache.h"

// my test implementation of search result caching, what I've got is
// a hash based on the query token, and a queue based on the time of search
// when a new token is searched for, it is first checked for in the hash, if it    
// exists, then the Gnut_List of shares is used
// otherwise, a new search is performed, this search result is then appended to 
// the hash and queue, with the oldest query being removed from the queue and the
// hash, much like in the routing section.  It would be a good idea upon a match to
// move the item back to the front of the list, so that it wouldn't expire as
// quickly if it was accessed frequently

Gnut_Hash *share_cache_hash=NULL;
Gnut_Queue *share_cache_queue=NULL;

uchar share_cache_hash_func(void *d)
{
  uchar b=0;
  int i;
  share_query_results *ss;
  
  ss=d;
  for (i=0;i<strlen(ss->query);i++) b+=ss->query[i];
  return b;
}

int share_cache_compare_func(void *a, void *b)
{
  share_query_results *ss,*ss2;
  ss=a;
  ss2=b;
  
  return strcmp(ss->query,ss2->query);
}

int share_cache_init()
{
  share_cache_hash=gnut_hash_new(share_cache_hash_func,share_cache_compare_func);
  share_cache_queue=gnut_queue_new();
  return 0;
}


Gnut_List *share_cache_run_search(char *token, int max)
{
  share_query_results ss;
  share_query_results *ss2;
  Gnut_List *gl;
  
  ss.query=token;
  
  ss2=gnut_hash_find(share_cache_hash,&ss);
  if (ss2) {
    // we hit a query in our cache, so we should return a copy of this
    // list, and possibly move this query up in the queue
    return share_copy_list(ss2->results);
  }
  // OK, we didn't have it in the hash, which means we need to
  // run the query, and then insert it into the cache, then return a copy
  
  gl=share_search(token,max);
  ss2=(share_query_results *) xmalloc(sizeof(share_query_results));
  ss2->query=strdup(token);
  ss2->results=share_copy_list(gl);
  
  gnut_hash_insert(share_cache_hash,ss2);
  gnut_queue_insert(share_cache_queue,ss2);
  
  return gl;
}

// take two share lists, and make a new list, not to exceed max num
// which is the logical AND of the two inputs.
Gnut_List *share_results_combine(Gnut_List *l1, Gnut_List *l2,int max)
{  
  int i=0;
  Gnut_List *tmp,*tmp2,*prev=NULL;
  Gnut_List *nl=NULL; 
  share_item *si1,*si2;
  share_item *nsi;
  
  tmp2=l2;
  
  for (tmp=l1;tmp;tmp=tmp->next) {
    si1=tmp->data;
    for (;tmp2;tmp2=tmp2->next) {
      si2=tmp2->data;
      if (si2->ref==si1->ref) {
        // they match, so we'll make a copy, and insert it into
        // the new list
        nsi=(share_item *) xmalloc(sizeof(share_item));
        memcpy(nsi,si2,sizeof(share_item));
        nsi->path=strdup(si2->path);
        nsi->upath=strdup(si2->upath);
        nsi->fpath=strdup(si2->fpath);
        nl=gnut_list_prepend(nl,nsi);
        
        if ((++i) > max) return nl;
        break;
      }
      if (si2->ref < si1->ref) {
        tmp2=prev;
        break;
      }
      prev=tmp2;
    }
  }
  return nl;
}
